
/*
   helgraf1.c
   
   Graphics routines for helicon.
   
   Author: Mark E. Nash
   Date:   June, 1988
*/

#include	<stdio.h>
#include	<stdlib.h>
#include	<graph.h>
#include	<malloc.h>
#include	<math.h>
#include	<string.h>
#include	<conio.h>
#include	<dos.h>
#include	"helimacs.h"
#include	"helimess.h"
#include	"helstruc.h"
#include	"helfdef.h"


/* externals */
extern char msg[];				/* in heliglob.h */
extern struct grafwindow *top_window;		/* in heliglob.h */
extern unsigned delcon;				/* in heliglob.h */
extern unsigned m1, m2, m3, m4;			/* in heliglob.h */
 


#pragma loop_opt(off)
	
struct grafwindow *open_grafwin(wtype, x1, y1, x2, y2, border_color, 
					background_color, cols, rows, list)
	unsigned char wtype;
	short x1, y1, x2, y2, border_color, background_color;
	unsigned short cols, rows;
	short list;
/*
   open_grafwin() sets up a window in a graphics mode.
   
   Entry: wtype is GRAFWIN for graphics window, TEXTWIN for text window, or
   		BWIN for a button window.
          (x1,y1) and (x2,y2) define the logical coordinates of the window,
          	which includes pixel-wide border.
          border_color is the color of the window border to use.
	  background_color is the color to make the window background.
	  If this is a text window, (cols,rows) is the size of the window
	  	in characters.
	  if list = TRUE, then add window to list and save background if
	  	needed, else this is an independent window.
	  
   Exit:  If no error, returns a grafwindow pointer,
                  and the window has been initialized,
          else, returns NULL.
*/
	{
	struct grafwindow *wp, *wtemp;	/* pointers to the window */
	short	xz1, yz1, xz2, yz2;		/* zoom coordinates */
	short	xzinc, yzinc;			/* x & y increment for zoom */
	unsigned delvar;			/* zoom delay counter */


	if (wtype == GRAFWIN)
/* check values */
		if (x2 <= (x1 + 2) || y2 <= (y1 + 2))
			return(NULL);
		else;
	else
		if (wtype == TEXTWIN || wtype == BWIN)
			{
/* check values */
			if (cols < 1 || cols > EMAXCOL || rows < 1 || 
								rows > EMAXROW)
				return(NULL);
				
/* set window size according to the number of character rows and columns */
			x2 = x1 + HPIXCHAR * (cols + 1);
			y2 = y1 + VPIXCHAR * (rows + 1);
			}
		else
			return(NULL);
	
/* get memory for the window structure */
	if ((wp = (struct grafwindow *)malloc(sizeof(struct grafwindow)))
								== NULL)
		{
		printf("\7");
		return(wp);
		}
/* update memory status */
	memupdt((unsigned long)sizeof(struct grafwindow), ALLOC);

	wp->list = list;
/* set top of list if needed */
	if (list && (top_window == NULL))
		top_window = wp;

	wp->ix1 = 0;
	wp->iy1 = 0;
	wp->size = 0l;
	wp->background = NULL;
	wp->next_window = NULL;

/* attach it to the end of the list */
	if (list && (wp != top_window))
		{
		wtemp = top_window;
		while (wtemp->next_window != NULL)
			wtemp = wtemp->next_window;
	
		wtemp->next_window = wp;
		}
	
	
/* type of window */
	wp->wintype = wtype;
	
/* turn off mouse cursor */
	MOUSE(2, 0, 0, 0);

/* If this is not the first window and is in the list, get the background. */
	if (wp != top_window && list)
		{
/* set up the background coordinates */
		wp->ix1 = x1;
		wp->iy1 = y1;
		
/* get the background */
		wp->size = _imagesize(x1, y1, x2, y2);
		if ((wp->background = (char far *)halloc(wp->size, 
								sizeof(char)))
								== NULL)
			{
			printf("\7");
/* turn mouse back on */
			MOUSE(1, 0, 0, 0);
			return(NULL);
			}
/* update memory status */
		memupdt((unsigned long)wp->size, ALLOC);

		_getimage(x1, y1, x2, y2, wp->background);
		}
	
/* zoom out border if text */
	if (wtype == TEXTWIN || wtype == BWIN)
		{
/* first set colors */
		_setfillmask(NULL);
		_setlinestyle(SOLID_LINE);

/* get the increments */
		xzinc = 1;
		yzinc = 1;
		if (x2 - x1 > y2 - y1)
			xzinc = (int)((float)(x2-x1) / (float)(y2-y1) + 0.5);
		else
			if (x2 - x1 < y2 - y1)
				yzinc = (int)((float)(y2-y1) / (float)(x2-x1)
									+ 0.5);
			
/* zoom it */
		xz1 = x1 + (x2-x1)/2;
		yz1 = y1 + (y2-y1)/2;
		xz2 = xz1;
		yz2 = yz1;
		_setcolor(border_color);
		if (wtype == TEXTWIN)
			_rectangle(_GBORDER, xz1, yz1, xz2, yz2);
		if (wtype == BWIN)
			_ellipse(_GBORDER, xz1, yz1, xz2, yz2);
		while (xz1-xzinc >= x1 && yz1-yzinc >= y1)
			{
			_setcolor(BLACK);
			if (wtype == TEXTWIN)
				_rectangle(_GBORDER, xz1, yz1, xz2, yz2);
			if (wtype == BWIN)
				_ellipse(_GBORDER, xz1, yz1, xz2, yz2);
		     	xz1 -= xzinc;
			yz1 -= yzinc;
			xz2 += xzinc;
			yz2 += yzinc;
			_setcolor(border_color);
			if (wtype == TEXTWIN)
				_rectangle(_GBORDER, xz1, yz1, xz2, yz2);
			if (wtype == BWIN)
				_ellipse(_GBORDER, xz1, yz1, xz2, yz2);
			for (delvar=0; delvar<delcon; delvar++);
			}
		}

/* draw window background and border */
	_setcolor(background_color);
	if (wtype == BWIN)
		_ellipse(_GFILLINTERIOR, x1, y1, x2, y2);
	else
		_rectangle(_GFILLINTERIOR, x1, y1, x2, y2);
	_setcolor(border_color);
	if (wtype == BWIN)
		_ellipse(_GBORDER, x1, y1, x2, y2);
	else
		_rectangle(_GBORDER, x1, y1, x2, y2);
	
	if (wtype == GRAFWIN)
		{
/* set up drawing area */
		wp->wbox.x1 = x1 + 1;
		wp->wbox.y1 = y1 + 1;
		wp->wbox.x2 = x2 - 1;
		wp->wbox.y2 = y2 - 1;
		}
/* 
   if a text window is specified, we want the entire window, including borders,
   for scaling....Therefore, it is a BAD IDEA to draw in a text window, since
   the graphics window drawing area does not include the borders.  Also, the
   graphics mode text window is automatically scaled, so scale_grafwindow
   should not be called for a text window.
*/
	if (wtype == TEXTWIN || wtype == BWIN)
		{
		wp->wbox.x1 = x1;
		wp->wbox.x2 = x2;
		wp->wbox.y1 = y1;
		wp->wbox.y2 = y2;
		scale_grafwin(wp, (double)-1.0, (double)-1.0,
				(double)cols, (double)rows);
		}

/* set up the character coord. maxima of text window */
	wp->maxcols = cols-1;
	wp->maxrows = rows-1;

/* set up initial cursor position */
	wp->curcol = 0;
	wp->currow = 0;

/* turn mouse back on */
	MOUSE(1, 0, 0, 0);
		
	return(wp);
	}

#pragma loop_opt()
	
	
	
int close_grafwin(wh)
	struct grafwindow **wh;
/*
   close_grafwin() closes the window referenced by the handle wh.
   
   Entry: wh is a handle to a window to close.
   
   Exit:  the window is closed and *wh = NULL;
          returns NOERR if successful, else ERR.
*/
	{
	struct grafwindow *wp, *rwp, *lwp;	/* window pointers */
	

/* turn off mouse cursor */
	MOUSE(2, 0, 0, 0);

	wp = *wh;
	if (wp != NULL)
		{
/* see if this is the top window */
		if (top_window == wp)
			{
/* erase the window */
			_setfillmask(NULL);
			_setlinestyle(SOLID_LINE);
			_setcolor(BLACK);
			if (_rectangle(_GFILLINTERIOR, wp->wbox.x1 - 1, 
				wp->wbox.y1 - 1, wp->wbox.x2 + 1, wp->wbox.y2 
								+ 1) == 0)
				printf("\7\7\7");
			top_window = NULL;
			}
		else
			if (wp->list == TRUE && wp != top_window)
				{
/* take it out of the list */
				rwp = top_window;
				lwp = NULL;
				while (rwp->next_window != NULL)
					{
					lwp = rwp;
					rwp = rwp->next_window;
					}
				lwp->next_window = NULL;
						
/* restore the image and free the memory */
				_putimage(wp->ix1, wp->iy1, wp->background, 
								_GPSET);
				hfree((void huge *)(wp->background));
/* update memory status */
				memupdt((unsigned long)wp->size, FRE);
				}

/* free window structure memory */
		free((void *)wp);
		memupdt((unsigned long)sizeof(struct grafwindow), FRE);
/* check the heap */
		switch (_heapchk())
			{
			case _HEAPOK:
			case _HEAPEMPTY:
				break;
			case _HEAPBADBEGIN:
			case _HEAPBADNODE:
				printf("\7");
			}
		}
		
	*wh = NULL;

/* turn mouse back on */
	MOUSE(1, 0, 0, 0);
		
	return(NOERR);
	}
	
	
	
int scale_grafwin(wp, xmin, ymin, xmax, ymax)
	struct grafwindow *wp;
	double xmin, ymin, xmax, ymax;
/*
   scale_grafwin() sets the scale of the window.
   
   Entry: wp points to the window for the scale.
          xmin, ymin, xmax, ymax define the scale of the window.
	  
   Exit:  The scale has been set if there were no errors, returns NOERR.
          Returns ERR if the scale is bad.
*/
	{
	if (wp == NULL)
		{
		sprintf(msg, W_NOPTR, "scale_grafwin()");
		err_window(msg);
		return(ERR);
		}
		
	if (xmin > xmax || ymin > ymax)
		return(ERR);
	
	wp->xsmin = xmin;
	wp->ysmin = ymin;
	wp->xsmax = xmax;
	wp->ysmax = ymax;
	
	wp->npx = fabs((wp->xsmax - wp->xsmin)/((double)(wp->wbox.x2 - 
								wp->wbox.x1)));
	wp->npy = fabs((wp->ysmax - wp->ysmin)/((double)(wp->wbox.y2 - 
								wp->wbox.y1)));
		
	return(NOERR);
	}

	
	
int win_dscale(wp, color, axis)
	struct grafwindow *wp;
	short color;
	short axis;
/*
   win_dscale() draws an axis scale inside the current window.
   
   Entry: wp is the current window, and has been set up and scaled.
          color is the color to draw the scale.
	  axis = XAXIS to draw x axis, else draw y axis.
	  
   Exit:  The x or y axis scale has been drawn.
*/
	{
	double	xstart, ystart;		/* starting major tick mark */
	double	xmajspace, ymajspace;	/* spacing between major tick marks */
	double	xminspace, yminspace;	/* spacing between minor tick marks */
	double	majctr, minctr;		/* for drawing tick marks */
	double	ytemp;			/* temp storage */
	int	lcol, lrow;		/* logical coord of tic label */
	

	if (wp == NULL)
		{
		sprintf(msg, W_NOPTR, "win_dscales()");
		err_window(msg);
		return(ERR);
		}
	
/* see if possible */
	if (wp->npx == 0.0 || wp->npy == 0.0)
		return(ERR);
		
/* get the tick marks using the CD tick mark algorithm */
	tickspace(wp->xsmin, wp->xsmax, &xstart, &xmajspace, &xminspace);
	tickspace(wp->ysmin, wp->ysmax, &ystart, &ymajspace, &yminspace);
	
	if (axis == XAXIS)
		{
/* draw the x-axis tick marks */
/* draw any x-axis minor tick marks before the start of major tick marks */
		for (minctr=xstart-xminspace; minctr>=wp->xsmin; 
							minctr-=xminspace)
			win_line(wp, minctr, wp->ysmin, minctr, 
			  wp->ysmin + (wp->ysmax - wp->ysmin) * MINTIC, color);
	
/* major tick mark loop */
		for (majctr=xstart; majctr<=wp->xsmax; majctr+=xmajspace)
			{
			win_line(wp, majctr, wp->ysmin, majctr, 
			  wp->ysmin + (wp->ysmax - wp->ysmin) * MAJTIC, color);
			sprintf(msg, "%g", majctr);
			lcol = wp->wbox.x1 + (int)((majctr - wp->xsmin) / 
				wp->npx) - (HPIXCHAR * (strlen(msg) / 2));

/* if label goes over either end, push it back on */
			if (wp->wbox.x2 < lcol + (int)(HPIXCHAR * strlen(msg))
								+ HPIXCHAR/2)
				lcol = wp->wbox.x2 - (int)(HPIXCHAR * 
						strlen(msg) - HPIXCHAR/2);
			if (wp->wbox.x1 > lcol - HPIXCHAR/2)
				lcol = wp->wbox.x1 + HPIXCHAR/2;
			
			lrow = wp->wbox.y1 + VPIXCHAR + (int)((wp->ysmax - 
						wp->ysmin) * MAJTIC / wp->npy);
			lrow = wp->wbox.y2 + wp->wbox.y1 - lrow;
			eputs_ega(msg, lcol, lrow, MESSLEN, SCLABCOLOR, FALSE);

/* nested minor tick mark loop */
			for (minctr=majctr+xminspace; minctr<(majctr+xmajspace)
				       && minctr<=wp->xsmax; minctr+=xminspace)
				win_line(wp, minctr, wp->ysmin, minctr,
					 wp->ysmin + (wp->ysmax - wp->ysmin) * 
								MINTIC, color);
			}
		}
	
	else
		{

/* draw the y-axis tick marks */
/* draw any y-axis minor tick marks before the start of major tick marks */
		for (minctr=ystart-yminspace; minctr>=wp->ysmin; 
							minctr-=yminspace)
			win_line(wp, wp->xsmin, minctr, wp->xsmin + 
			      (wp->xsmax - wp->xsmin) * MINTIC, minctr, color);
	
/* major tick mark loop */
		for (majctr=ystart; majctr<=wp->ysmax; majctr+=ymajspace)
			{
			win_line(wp, wp->xsmin, majctr, wp->xsmin + 
			      (wp->xsmax - wp->xsmin) * MAJTIC, majctr, color);

			sprintf(msg, "%g", majctr);
			lrow = wp->wbox.y1 + (int)((majctr - wp->ysmin) / 
								wp->npy);
			lrow = wp->wbox.y2 + wp->wbox.y1 - lrow;

/* if label goes over either end, push it back on */
			if (wp->wbox.y2 < lrow + (int)(VPIXCHAR * strlen(msg))
								+ VPIXCHAR/2)
				lrow = wp->wbox.y2 - (int)(VPIXCHAR * 
						strlen(msg) - VPIXCHAR/2);
			if (wp->wbox.y1 > lrow - VPIXCHAR/2)
				lrow = wp->wbox.y1 + VPIXCHAR/2;
			
			lcol = wp->wbox.x1 + HPIXCHAR + (int)((wp->xsmax - 
						wp->xsmin) * MAJTIC / wp->npx);
			eputs_ega(msg, lcol, lrow, MESSLEN, SCLABCOLOR, FALSE);

/* nested minor tick mark loop */
			for (minctr=majctr+yminspace; minctr<(majctr+ymajspace)
				       && minctr<=wp->ysmax; minctr+=yminspace)
				win_line(wp, wp->xsmin, minctr, wp->xsmin + 
					(wp->xsmax - wp->xsmin) * MINTIC, 
								minctr, color);
			}
		}

	return(NOERR);
	}



int win_line(wp, x1, y1, x2, y2, color)
	struct grafwindow *wp;
	double x1, y1, x2, y2;
	short color;
/*
   win_line() draws a line in a window.
   
   Entry: wp points to the current window.
          (x1,y1) is where the line starts.
	  (x2,y2) is where the line ends.
	  color is the color to make the line.
	  
   Exit:  The portion of the line lying in the window is drawn.
          Returns NOERR.
*/
	{
	short int xl1, yl1, xl2, yl2;		/* logical line */
	

	if (wp == NULL)
		{
		sprintf(msg, W_NOPTR, "win_line()");
		err_window(msg);
		return(ERR);
		}
		
/* clip the line */
	if (clip(wp, &x1, &y1, &x2, &y2) == NOPLOT)
		return(NOERR);

/* scale the line */
	y1 = wp->ysmax + wp->ysmin - y1;
	y2 = wp->ysmax + wp->ysmin - y2;
	if (wp->npx != 0.0 && wp->npy != 0.0)
		{
		xl1 = (short)(((x1 - wp->xsmin) / wp->npx) + .5) + wp->wbox.x1;
		yl1 = (short)(((y1 - wp->ysmin) / wp->npy) + .5) + wp->wbox.y1;
		xl2 = (short)(((x2 - wp->xsmin) / wp->npx) + .5) + wp->wbox.x1;
		yl2 = (short)(((y2 - wp->ysmin) / wp->npy) + .5) + wp->wbox.y1;

/* turn off mouse cursor */
		MOUSE(2, 0, 0, 0);

/* draw the line */
		_setcolor(color);
		_moveto(xl1, yl1);
		_lineto(xl2, yl2);
	
/* turn mouse back on */
		MOUSE(1, 0, 0, 0);
		}
		
	return(NOERR);
	}
	
	
	
int win_ell(wp, x1, y1, x2, y2, color)
	struct grafwindow *wp;
	double x1, y1, x2, y2;
	short color;
/*
   win_ell() draws an ellipse in a window.
   
   Entry: wp points to the current window.
          (x1,y1) is the upper left corner of the rectangle outlining the ell.
	  (x2,y2) is the lower right corner of the rect. outlining the ell.
	  color is the color to make the ellipse.
	  
   Exit:  The ellipse is drawn in the window with NO clipping.
          Returns NOERR.
*/
	{
	short int xl1, yl1, xl2, yl2;		/* logical rectangle */
	

	if (wp == NULL)
		{
		sprintf(msg, W_NOPTR, "win_ell()");
		err_window(msg);
		return(ERR);
		}
		
/* scale the ellipse */
	y1 = wp->ysmax + wp->ysmin - y1;
	y2 = wp->ysmax + wp->ysmin - y2;
	if (wp->npx != 0.0 && wp->npy != 0.0)
		{
		xl1 = (short)(((x1 - wp->xsmin) / wp->npx) + .5) + wp->wbox.x1;
		yl1 = (short)(((y1 - wp->ysmin) / wp->npy) + .5) + wp->wbox.y1;
		xl2 = (short)(((x2 - wp->xsmin) / wp->npx) + .5) + wp->wbox.x1;
		yl2 = (short)(((y2 - wp->ysmin) / wp->npy) + .5) + wp->wbox.y1;

/* turn off mouse cursor */
		MOUSE(2, 0, 0, 0);

/* draw the ellipse */
		_setcolor(color);
		_ellipse(_GBORDER, xl1, yl1, xl2, yl2);
	
/* turn mouse back on */
		MOUSE(1, 0, 0, 0);
		}
		
	return(NOERR);
	}
	
	
	
int win_axes(wp, color)
	struct grafwindow *wp;
	short color;
/*
   win_axes() draws the coordinate axes in the window.
   
   Entry: wp points to the current window.
	  color is the color of the axes.
	  
   Exit:  If axes are within the window, they are drawn with the given color.
          Returns NOERR.
*/
	{
	if (wp == NULL)
		{
		sprintf(msg, W_NOPTR, "win_axes()");
		err_window(msg);
		return(ERR);
		}
		
	win_line(wp, 0.0, wp->ysmin, 0.0, wp->ysmax, color);
	win_line(wp, wp->xsmin, 0.0, wp->xsmax, 0.0, color);
	return(NOERR);
	}

	
	
	
int win_point(wp, x, y, color)
	struct grafwindow *wp;
	double x, y;
	short color;
/*
   win_point() sets a point in a window to a given color.
   
   Entry: wp points to the current window.
          (x,y) is the coordinate of the point.
	  color is the color of the point.
	  
   Exit:  If the point is within the window, it is set to the given color.
          Returns NOERR.
*/
	{
	short int xl, yl;		/* logical point */
	

	if (wp == NULL)
		{
		sprintf(msg, W_NOPTR, "win_point()");
		err_window(msg);
		return(ERR);
		}
		
/* clip the point */
	if (x < wp->xsmin || x > wp->xsmax || y < wp->ysmin || y > wp->ysmax)
		return(NOERR);

/* scale the point */
	y = wp->ysmax + wp->ysmin - y;
	if (wp->npx != 0.0 && wp->npy != 0.0)
		{
		xl = (short)(((x - wp->xsmin) / wp->npx) + .5) + wp->wbox.x1;
		yl = (short)(((y - wp->ysmin) / wp->npy) + .5) + wp->wbox.y1;

/* turn off mouse cursor */
		MOUSE(2, 0, 0, 0);

/* plot the point */
		_setcolor(color);
		_setpixel(xl, yl);
	
/* turn mouse back on */
		MOUSE(1, 0, 0, 0);
		}
		
	return(NOERR);
	}
	
	
	
int win_rect(wp, x1, y1, x2, y2, color, fill)
	struct grafwindow *wp;
	double	x1, y1, x2, y2;
	short	color;
	short	fill;
/*
   win_rect() draws a rectangle in the given window.
   
   Entry: wp points to the window to use.
          (x1,y1) is the upper left corner of the rectangle.
	  (x2,y2) is the lower right corner of the rectangle.
	  color is the color to make the rectangle.
	  fill = TRUE for fill with given color.
	  
   Exit:  The specified rectangle has been drawn & NOERR returned.
*/
	{
	short	xl, yl;		/* for floodfill */
	double	xp, yp;		/* for floodfill */
	
	
	if (wp == NULL)
		{
		sprintf(msg, W_NOPTR, "win_rect()");
		err_window(msg);
		return(ERR);
		}

	win_line(wp, x1, y1, x1, y2, color);
	win_line(wp, x1, y2, x2, y2, color);
	win_line(wp, x2, y2, x2, y1, color);
	win_line(wp, x2, y1, x1, y1, color);
	
/* fill if necessary */
	if (fill && (wp->npx != 0.0 && wp->npy != 0.0))
		{
		xp = fabs(x2 - x1) / 2.0 + x1;
		yp = fabs(y2 - y1) / 2.0 + y1;
/* scale a point */
		yp = wp->ysmax + wp->ysmin - yp;
		xl = (short)(((xp - wp->xsmin) / wp->npx) + .5) + wp->wbox.x1;
		yl = (short)(((yp - wp->ysmin) / wp->npy) + .5) + wp->wbox.y1;
/* turn off mouse cursor */
		MOUSE(2, 0, 0, 0);
		_setfillmask(NULL);
		_floodfill(xl, yl, color);
/* turn mouse back on */
		MOUSE(1, 0, 0, 0);
		}
	
	return(NOERR);
	}
	
	
	
int win_func(wp, ftype, pa, color)
	struct grafwindow *wp;
	unsigned ftype;
	double pa[MAXPARS+1][NUMDIST+1];
	short	color;
/*
   win_func() draws the given function in the given window.
   
   Entry: wp is the window to use.
          ftype is the function type.
	  pa[][] is the parameter array.
	  color is the color to draw the function.
	  
   Exit:  If no error, the function has been drawn, returns NOERR
          Else, returns ERR.
*/
	{
	double	x1, y1, x2, y2;		/* line coordinates */
	double	xmax, xinc;		/* maximum and increment values */
	int	lv1, lv2;		/* for checking legal values */
	

	if (wp == NULL)
		{
		sprintf(msg, W_NOPTR, "win_func()");
		err_window(msg);
		return(ERR);
		}
		
/* see if possible */
	if (wp->npx == 0.0 || wp->npy == 0.0)
		return(ERR);
		
/* initialize these for speed */
	xmax = wp->xsmax;
	if (ftype < FIRSTDISC)
		{
		xinc = wp->npx;
		x1 = wp->xsmin;
		}
	else
		{
		xinc = 1.0;
		x1 = (double)((int)wp->xsmin);
		}

/* loop and draw - get the first legal values.  I am assuming that there are
   no undefined values between the endpoints of the line. */
	x2 = x1 + xinc;
	lv1 = fval(ftype, x1, pa, FALSE, &y1);
	lv2 = fval(ftype, x1, pa, FALSE, &y2);
	while (x1 < xmax)
		{
		if (ftype < FIRSTDISC)
			if (lv1 && lv2)
				win_line(wp, x1, y1, x2, y2, color);
			else;
		else
			if (lv1)
				win_line(wp, x1-0.1, y1, x1+0.1, y1, color);
			else;
			
		x1 = x2;
		x2 = x1 + xinc;
		lv1 = lv2;
		y1 = y2;
		lv2 = fval(ftype, x2, pa, FALSE, &y2);
		}
	
	return(NOERR);
	}



int clip(wp, xc1, yc1, xc2, yc2)
	struct grafwindow *wp;
	double *xc1, *yc1, *xc2, *yc2;
/*
   clip() clips a line to fit in a window.
   
   Entry: wp is the scaled window to clip the line to.
	  (*xc1,*yc1) is one end of the line.
	  (*xc2,*yc2) is the other end of the line.
	  
   Exit:  xc1, yc1, xc2, yc2 have been updated fit in the clipping window.
*/
	{
	double	x1, y1, x2, y2;			/* line endpoints */
	double	m, xt1, yt1, xt2, yt2;		/* for clipping calculation */
	unsigned vert, horiz;			/* special cases */
	short	top, bottom, left, right;	/* valid intercept flags */


	if (wp == NULL)
		{
		sprintf(msg, W_NOPTR, "clip()");
		err_window(msg);
		return(ERR);
		}
		
/* get line values */
	x1 = *xc1;
	y1 = *yc1;
	x2 = *xc2;
	y2 = *yc2;
	
/* check to see if line is outside window with endpoints in the same sector */
	if (x1 < wp->xsmin && x2 < wp->xsmin)
		return(NOPLOT);
	if (x1 > wp->xsmax && x2 > wp->xsmax)
		return(NOPLOT);
	if (y1 < wp->ysmin && y2 < wp->ysmin)
		return(NOPLOT);
	if (y1 > wp->ysmax && y2 > wp->ysmax)
		return(NOPLOT);

/* init special case variables */
	vert = FALSE;
	horiz = FALSE;
	if (fabs(y2-y1) <= wp->npy)
		{
		horiz = TRUE;

/* see if we need to draw the horizontal line */
		if (y1 < wp->ysmin || y1 > wp->ysmax)
			return(NOPLOT);
		}
	else
		if (fabs(x2-x1) <= wp->npx)
			{
			vert = TRUE;
			
/* see if we need to draw the vertical line */
			if (x1 < wp->xsmin || x1 > wp->xsmax)
				return(NOPLOT);
			}

/* initialize intercept flags */
	left = FALSE;
	right = FALSE;
	top = FALSE;
	bottom = FALSE;
	if (horiz == FALSE && vert == FALSE)
		{
		m = (y2 - y1) / (x2 - x1);
/* find theoretical intercepts */
		yt1 = m * wp->xsmin - m*x1 + y1;    /* y int of left side */
		yt2 = m * wp->xsmax - m*x1 + y1;    /* y int of right side */
		xt1 = wp->ysmin / m - y1/m + x1;    /* x int of bottom side */
		xt2 = wp->ysmax / m - y1/m + x1;    /* x int of top side */

/* find out which theoretical intercepts are valid */
		if (yt1 < wp->ysmax && yt1 > wp->ysmin)
			left = TRUE;
		if (yt2 < wp->ysmax && yt2 > wp->ysmin)
			right = TRUE;
		if (xt1 <= wp->xsmax && xt1 >= wp->xsmin)
			bottom = TRUE;
		if (xt2 <= wp->xsmax && xt2 >= wp->xsmin)
			top = TRUE;

/* check to see if line is in window */
		if (left == FALSE && right == FALSE && bottom == FALSE &&
								top == FALSE)
			return(NOPLOT);
		}
	else
		if (horiz == TRUE)
			{
			yt1 = y1;	/* y int of left side */
			yt2 = y2;	/* y int of right side */
			right = TRUE;
			left = TRUE;
			}
		else	/* vertical == TRUE */
			{
			xt1 = x1;	/* x int of bottom side */
			xt2 = x2;	/* x int of top side */
			top = TRUE;
			bottom = TRUE;
			}

/* check ranges and assign new values if needed */
/* clip left */
	if (left == TRUE)
		if (x1 < wp->xsmin)
			{
			x1 = wp->xsmin;
			y1 = yt1;
			}
		else
			if (x2 < wp->xsmin)
				{
				x2 = wp->xsmin;
				y2 = yt1;
				}
/* clip right */
	if (right == TRUE)
		if (x1 > wp->xsmax)
			{
			x1 = wp->xsmax;
			y1 = yt2;
			}
		else
			if (x2 > wp->xsmax)
				{
				x2 = wp->xsmax;
				y2 = yt2;
				}
/* clip bottom */
	if (bottom == TRUE)
		if (y1 < wp->ysmin)
			{
			x1 = xt1;
			y1 = wp->ysmin;
			}
		else
			if (y2 < wp->ysmin)
				{
				x2 = xt1;
				y2 = wp->ysmin;
				}
/* clip top */
	if (top == TRUE)
		if (y1 > wp->ysmax)
			{
			x1 = xt2;
			y1 = wp->ysmax;
			}
		else
			if (y2 > wp->ysmax)
				{
				x2 = xt2;
				y2 = wp->ysmax;
				}
					
/* set up return values */
	*xc1 = x1;
	*yc1 = y1;
	*xc2 = x2;
	*yc2 = y2;
	return(PLOT);
	}
	
	
	
int win_clear(wp, color)
	struct grafwindow *wp;
	short	color;
/*
   win_clear() clears a given window to a given color.
   
   Entry: wp points to the window to use.
	  color is the color to make the window.
	  
   Exit:  The specified has been cleared to the given color.
*/
	{
	short	x1, y1, x2, y2;		/* for drawing rectangle */
	
	
	if (wp == NULL)
		return(ERR);
		
/* set clear area */
	x1 = wp->wbox.x1;
	y1 = wp->wbox.y1;
	x2 = wp->wbox.x2;
	y2 = wp->wbox.y2;
	
/* reset if text window */
	if (wp->wintype == TEXTWIN || wp->wintype == BWIN)
		{
		x1++;
		y1++;
		x2--;
		y2--;
		}
		
/* turn off mouse cursor */
	MOUSE(2, 0, 0, 0);

	_setcolor(color);
	if (wp->wintype == BWIN)
		_ellipse(_GFILLINTERIOR, x1, y1, x2, y2);
	else
		_rectangle(_GFILLINTERIOR, x1, y1, x2, y2);

/* turn mouse back on */
 	MOUSE(1, 0, 0, 0);
		
	return(NOERR);
	}
	
	
	
int getpt(col, row)
	short int col, row;
/*
   getpt() gets the color of a point on the ega graphics screen.
   
   Entry: row, col is the position of the point in physical coords.
   
   Exit:  Returns the color of the point.
*/
	{
	short	 color;				/* the value read */
	union REGS regs;			/* for talking to BIOS */


	regs.h.ah = 0x0D;
	regs.h.bh = 0;
	regs.x.dx = row;
	regs.x.cx = col;
	int86(0x10, &regs, &regs);
	color = (short)regs.h.al;

	return(color);
	}
	
	
		
int ctob(color)
	short int color;
/*
   ctob() converts a color pixel value to either black or white.
*/
	{
	int	bw;		/* black/white value */
	
	
	bw = 1;			/* white is default */
	if (color == BLACK || color == LIGHTRED)
		bw = 0;

	return(bw);
	}
